import os
import subprocess
import tempfile

from tempdir import TempDir


class LaTeXError(Exception):
    """LaTeX call exception."""

    #: Contents of the log-file that was generated (if any).
    log = None

    #: The :class:`subprocess.CalledProcessError` that was thrown originally.
    call_exc = None

    def __str__(self):
        return str(self.log)

    @classmethod
    def from_proc_error(cls, e, logfn=None):
        err = cls()
        err.log = None
        err.call_exc = e

        if os.path.exists(logfn):
            err.log = open(logfn).read()

        return err


def build_pdf(source,
              latex_cmd='pdflatex',
              texinputs=[''],
              max_runs=15):
    """Generates a PDF from LaTeX a source, using pdflatex and a temporary
    directory, to avoid leaving behind temporary files.

    pdflatex will be run as many times as necessary; at ``max_runs`` runs, an
    exception is raised and the generation is aborted with a
    :class:`RuntimeError`.

    If there are errors calling pdflatex, a :class:`~latex.build.LaTeXError` is
    raised.

    :param source: The LaTeX source.
    :param latex_cmd: The path for the ``pdflatex`` binary to use.
    :param texinputs: Include paths for TeX. An empty string causes the default
                      path to be added (see the tex manpage).
    :param max_runs: Maximum number of reruns.
    :returns: A binary string of the generated PDF file.
    """

    with TempDir() as tmpdir:
        input_fd, input_fn = tempfile.mkstemp(suffix='.latex', dir=tmpdir)
        input_file = os.fdopen(input_fd, 'w')

        # write the source file
        input_file.write(source)
        input_file.close()

        # calculate output filename
        base_fn = os.path.splitext(input_fn)[0]
        output_fn = base_fn + '.pdf'
        aux_fn = base_fn + '.aux'
        args = [latex_cmd,
                '-interaction=batchmode',
                '-halt-on-error',
                '-no-shell-escape',
                '-file-line-error',
                input_fn]

        # create environment
        newenv = os.environ.copy()
        newenv['TEXINPUTS'] = os.pathsep.join(texinputs) + os.pathsep

        prev_aux = None
        while True:
            try:
                subprocess.check_call(
                    args,
                    cwd=tmpdir,
                    env=newenv,
                    stdin=open(os.devnull, 'r'),
                    stdout=open(os.devnull, 'w'),
                )
            except subprocess.CalledProcessError as e:
                raise LaTeXError.from_proc_error(e, base_fn + '.log')

            # check aux-file
            aux = open(aux_fn, 'rb').read()

            if aux == prev_aux:
                break

            prev_aux = aux

            max_runs -= 1

            if max_runs == 0:
                raise RuntimeError(
                    'Maximum number of runs ({}) without a stable .aux file '
                    'reached.'.format(max_runs)
                )

        # read generated .pdf
        return open(output_fn, 'rb').read()

        # FIXME: http://vim-latex.sourceforge.net/documentation/latex-suite
        #        /compiling-multiple.html



import re

CHAR_ESCAPE = {
    u'&': u'\\&',
    u'%': u'\\%',
    u'$': u'\\$',
    u'#': u'\\#',
    u'_': u'\\_',
    u'{': u'\\{',
    u'}': u'\\}',
    u'~': u'\\textasciitilde{}',
    u'^': u'\\textasciicircum{}',
    u'\\': u'\\textbackslash{}',

    # these may be optional:
    u'<': u'\\textless{}',
    u'>': u'\\textgreater{}',
    u'|': u'\\textbar{}',
    u'"': u'\\textquoteddbl{}',
}


def _sub_tbl(tbl):
    return r'|'.join(re.escape(k) for k in sorted(tbl.keys()))


ESCAPE_RE = re.compile(r'\n+|' + _sub_tbl(CHAR_ESCAPE))


def escape(s, fold_newlines=True):
    """Escapes a string to make it usable in LaTeX text mode. Will replace
    special characters as well as newlines.

    :param s: The string to escape.
    :param fold_newlines: If true, multiple newlines will be reduced to just a
                          single ``\\``. Otherwise, whitespace is kept intact
                          by adding multiple ``[n\baselineskip]``.
    """
    def sub(m):
        c = m.group()
        if c in CHAR_ESCAPE:
            return CHAR_ESCAPE[c]

        if c.isspace():
            if fold_newlines:
                return r'\\'
            return r'\\[{}\baselineskip]'.format(len(c))

    return ESCAPE_RE.sub(sub, s)



from __future__ import absolute_import

from markupsafe import Markup
from jinja2 import Environment

from . import escape


class LatexMarkup(Markup):
    def unescape(self):
        raise NotImplementedError

    def stripstags(self):
        raise NotImplementedError

    @classmethod
    def escape(cls, s):
        if hasattr(s, '__html__'):
            return s.__html__()

        rv = escape(s)
        if rv.__class__ is not cls:
            return cls(rv)
        return rv


ENV_ARGS = {
    'block_start_string': '\BLOCK{',
    'block_end_string': '}',
    'variable_start_string': '\VAR{',
    'variable_end_string': '}',
    'comment_start_string': '\#{',
    'comment_end_string': '}',
    'line_statement_prefix': '%-',
    'line_comment_prefix': '%#',
    'trim_blocks': True,
    'autoescape': False,
}


def make_env(*args, **kwargs):
    """Creates an :py:class:`~jinja2.Environment` with different defaults.

    Per default, ``autoescape`` will be disabled and ``trim_blocks`` enabled.
    All start/end/prefix strings will be changed for a more LaTeX-friendly
    version (see the docs for details).

    Any arguments will be passed on to the :py:class:`~jinja2.Environment`
    constructor and override new values.

    Finally, the ``|e``, ``|escape`` and ``|forceescape`` filters will be
    replaced with a call to :func:`latex.escape`."""
    ka = ENV_ARGS.copy()
    ka.update(kwargs)

    env = Environment(*args, **ka)
    env.filters['e'] = LatexMarkup.escape
    env.filters['escape'] = LatexMarkup.escape
    env.filters['forceescape'] = LatexMarkup.escape  # FIXME: this is a bug
    return env
